"""Base constructs for build package objects
"""
import re
import string
from typing import List, Optional, Iterator, Dict
from collections.abc import Mapping
from collections import Counter
from dataclasses import dataclass, field, make_dataclass
from dataclasses_json import dataclass_json, config, Undefined, CatchAll
from bgameb.errors import (
    ComponentNameError, ComponentClassError, ComponentIdError
        )
from loguru import logger


logger.disable('bgameb')


def log_enable(
    log_path: str = './logs/game.log',
    log_level: str = 'DEBUG'
        ) -> None:
    """Enable logging

    Args:
        log_path (str, optional): path to log file.
                                  Defaults to './logs/game.log'.
        log_level (str, optional): logging level. Defaults to 'DEBUG'.
    """
    logger.remove()
    logger.add(
        sink=log_path,
        level=log_level,
        format='{extra[classname]}: "{extra[name]}" -> func {function} | ' +
        '{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}',
    )
    logger.enable('bgameb')


@dataclass_json
@dataclass(repr=False)
class Component(Mapping):
    """Component mapping
    """
    def __init__(
        self,
        *args,
        **kwargs
            ) -> None:
        """Args must be a dicts
        """
        super().__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                self.__dict__.update(arg)
            else:
                raise AttributeError('Args must be a dicts')
        if kwargs:
            self.__dict__.update(kwargs)

    @property
    def inclusion(self) -> Dict[str, str]:
        return {
            k: v for k, v
            in self.__dict__.items()
            if not k.startswith('_')
            and not k.startswith('current')
                }

    def __iter__(self) -> Iterator:
        return iter(self.__dict__)

    def __getattr__(self, attr: str):
        try:
            return self.__dict__[attr]
        except KeyError:
            raise AttributeError(attr)

    def __delattr__(self, attr: str) -> None:
        try:
            del self.__dict__[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setitem__(self, attr: str, value):
        attr = self._make_name(attr)
        self.__dict__.update({attr: value})

    def __getitem__(self, attr: str):
        return self.__dict__[attr]

    def __delitem__(self, attr: str) -> None:
        del self.__dict__[attr]

    def __repr__(self) -> str:
        items = {f"{k}={v!r}" for k, v in self.inclusion.items()}
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __len__(self) -> int:
        return len(self.inclusion)

    def _is_unique(self, name: str) -> Optional[bool]:
        """Chek is name of nested component is unique
        for based component __dict__

        Args:
            name (str): name of component

        Raises:
            ComponentNameError: name not unique

        Returns:
            True: is unique
        """
        if name not in self.__dict__.keys():
            return True
        raise ComponentNameError(name)

    def _is_valid(self, name: str) -> bool:
        """Chek is name of component contains correct symbols
        match [a-zA-Z_][a-zA-Z0-9_]*$ expression:

            * a-z, A-Z, 0-9 symbols
            * first letter not a number amd not a _
            * can be used _ symbol in subsequent symbols

        Args:
            name (str): name of component

        Raises:
            ComponentNameError: name is not valid

        Returns:
            bool: is valid
        """
        if re.match("[a-z][a-z0-9_]*$", str(name)):
            return True
        return False

    def _make_name(self, name: str) -> str:
        """
        Replace spaces and other specific characters
        in the name with _

        Args:
            name (str): name of component

        Returns:
            name (str): safe name of component
        """
        name = str(name).lower()
        available = set(string.ascii_letters.lower() + string.digits + '_')

        if " " in name:
            name = name.replace(' ', '_')

        diff = set(name).difference(available)
        if diff:
            for char in diff:
                name = name.replace(char, '_')

        if not self._is_valid(name):
            raise ComponentNameError(name)

        return name

    def _update(
        self,
        component,
            ) -> None:
        """Update Component dict with safe name

        Args:
            component: component instance
        """
        name = self._make_name(component.id)

        if self._is_unique(name):
            comp = component.__class__(**component.to_dict())

        if name not in self.__dataclass_fields__.keys():
            self.__class__ = make_dataclass(
                self.__class__.__name__,
                fields=[(name, type(comp), field(default=comp))],
                bases=(self.__class__, ),
                repr=False
                )

        self.__dict__.update({name: comp})

    def get_names(self) -> List[str]:
        """Get names of all components in Component

        Returns:
            List[str]: list of components names
        """
        return [name for name in self.inclusion]


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass(repr=False)
class Base(Component):
    """Base class for game, stuff, tools players and other components

    Attr:
        - id (str): id of component
        - other (Dict[str, Any]): all other data, added to instance
                                  at declaration
        - counter (Counter): counter object
        - _type (Optional[str]): type for check when this component
          can be added
        - _types_to_add (List[str]): types of components, that can
          be added

    Counter is a `collection.Counter
    <https://docs.python.org/3/library/collections.html#collections.Counter>`_
    """
    id: str
    other: CatchAll = field(default_factory=dict)
    counter: Counter = field(default_factory=dict)  # type: ignore
    _type: Optional[str] = field(
        default=None,
        metadata=config(exclude=lambda x: True),  # type: ignore
        repr=False,
        )
    _types_to_add: List[str] = field(
        default_factory=list,
        metadata=config(exclude=lambda x: True),  # type: ignore
        repr=False,
            )

    def __post_init__(self) -> None:
        # check id
        if not isinstance(self.id, str):
            raise ComponentIdError(self.id)

        # int counter
        self.counter = Counter()

        # set self_type
        self._type = self.__class__.__name__.lower()

        # set logger
        self._logger = logger.bind(
            classname=self.__class__.__name__,
            name=self.id)
        if self._type == 'game':
            self._logger.info('===========NEW GAME============')
        self._logger.info(
            f'{self.__class__.__name__} created with id="{self.id}".'
            )

    def add(self, component) -> None:
        """Add another component to this component

        Args:
            component (Component): component instance
        """
        if component._type in self._types_to_add:
            self._update(component)
            self._logger.info(f'"{component.id}" is added to "{self.id}".')
        else:
            raise ComponentClassError(component, self._logger)



"""Main engine to create game
"""
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from bgameb.base import Base, log_enable
from bgameb.types import COMPONENTS


@dataclass_json
@dataclass(repr=False)
class Game(Base):
    """The main game object
    """

    def __post_init__(self) -> None:
        super().__post_init__()
        self._types_to_add = COMPONENTS


if __name__ == '__main__':
    import bgameb
    from pprint import pprint
    log_enable()

    game = bgameb.Game('one board game')
    game.add(bgameb.Steps('Game Steps'))
    game.game_steps.add(bgameb.Step('step 0'))
    game.game_steps.add(bgameb.Step('step 1', priority=1))

    game.add(bgameb.Deck('Cards Deck'))
    card = bgameb.Card('One card', description='story')  # type: ignore
    game.cards_deck.add(card)
    game.cards_deck.one_card.count = 3

    print(f'Other: {game.cards_deck.one_card.other}')
    game.cards_deck.one_card.counter['yellow'] = 12
    game.cards_deck.one_card.counter['banana'] = 0

    steps = game.game_steps.deal()
    game.cards_deck.deal()
    deck = game.cards_deck.shuffle()

    game.IS_ACTIVE = True  # type: ignore

    game.add(bgameb.Shaker('blue shaker'))
    game.blue_shaker.add(
        bgameb.Dice('dice#8', sides=8, count=10)
            )

    result = game.blue_shaker.dice_8.roll()

    print(f'Result of shke: {result}')
    print('='*20 + '\n')
    print(f'Repr: {repr(game)}')
    print('='*20 + '\n')
    print(f'Print: {game}')
    print('='*20 + '\n')
    pprint(game.to_dict())
    print('='*20 + '\n')
    pprint(deck)
    print('='*20 + '\n')
    print(game.cards_deck.get_current_names())
    print('='*20 + '\n')
    print(f' Get_ids(): {game.get_names()}')
    print('='*20 + '\n')
    pprint(f'Len: {len(game)}')
    print('='*20 + '\n')
    print(f'Dir: {dir(game)}')



"""Game tools classes
"""
import random
from collections import deque
from heapq import heappop, heappush
from typing import Tuple, Dict, List, Deque, Optional
from dataclasses import dataclass, field, replace
from dataclasses_json import config, dataclass_json
from bgameb.base import Base
from bgameb.markers import Step
from bgameb.items import Card, Dice
from bgameb.errors import ArrangeIndexError
from bgameb.types import MARKERS_ITEMS


@dataclass_json
@dataclass(repr=False)
class BaseTool(Base):
    """Base class for game tools (like decks or shakers)
    """

    def __post_init__(self) -> None:
        super().__post_init__()
        self._types_to_add = MARKERS_ITEMS


@dataclass_json
@dataclass(repr=False)
class Shaker(BaseTool):
    """Create shaker for roll dices or flip coins
    """

    def __post_init__(self) -> None:
        super().__post_init__()

    def roll(self) -> Dict[str, Tuple[int]]:
        """Roll all stuff in shaker and return results

        Return:
            Dict[str, Tuple[int]]: result of roll

        .. code-block::
            :caption: Example:

                {
                    "six_dice": (5, 3, 2, 5),
                    "twenty_dice": {2, 12, 4},
                }
        """
        roll = {}

        for comp in self:
            if isinstance(self[comp], Dice):
                roll[self[comp].id] = self[comp].roll()

        self._logger.debug(f'Result of roll: {roll}')

        return roll


@dataclass_json
@dataclass(repr=False)
class Deck(BaseTool):
    """Deck object

    Deck ia a Bag subclass that contains Cards for
    define curent game deque.

    You can add cards, define it counts and deal a deck.
    Result is saved in current deck attr as deque object. This object
    has all methods of
    `python deque
    <https://docs.python.org/3/library/collections.html#deque-objects>`_

    .. code-block::
        :caption: Example:

            deque(Card1, Card3, Card2, Card4)
    """
    current: Deque[Card] = field(
        default_factory=deque,
        metadata=config(exclude=lambda x: True),  # type: ignore
        repr=False,
        )

    def __post_init__(self) -> None:
        super().__post_init__()

    def _card_replace(self, card: Card) -> None:
        """Replace card in current
        """
        replaced = replace(card)
        del replaced.count
        self.current.append(replaced)

    def deal(self, source: Optional[List[str]] = None) -> Deque[Card]:
        """Deal new random shuffled deck and save it to
        self.current

        Args:
            source (Optional[List[str]]): list of stuff ids
        """
        self.current.clear()

        if not source:

            for comp in self:
                if isinstance(self[comp], Card):
                    for _ in range(self[comp].count):
                        self._card_replace(self[comp])

        else:

            names = self.get_names()
            used: Dict[str, Card] = {}

            for name in source:
                if name in names:
                    if not used.get(name):
                        for comp in self:
                            if isinstance(self[comp], Card) \
                                    and self[comp].id == name:
                                self._card_replace(self[comp])
                                used[name] = self[comp]
                                break
                    else:
                        self._card_replace(self[comp])

        self._logger.debug(f'Is deal cards: {self.current}')
        return self.current

    def shuffle(self) -> Deque[Card]:
        """Random shuffle current deck
        """
        random.shuffle(self.current)
        self._logger.debug(f'Is shuffled: {self.current}')
        return self.current

    def to_arrange(
        self,
        start: int,
        end: int
            ) -> Tuple[
                List[Card],
                Tuple[List[Card], List[Card]]
                    ]:
        """Prepare current deck to arrange

        Args:
            start (int): start of slice
            end (int): end of slice

        Start and end cant be less than 0 and end must be greater than start.
        Arranged deck are splited to two part - center (used for arrange part
        of deck) and tupple with last part of deck. You can rearrange center
        part and then concatenate it with last part in new deque by arrange()
        method.

        Return:
            Tuple[List[Card], Tuple[List[Card]]: parts to arrange
        """
        if start < 0 or end < 0 or end < start:
            raise ArrangeIndexError(
                message=f'Nonpositive or broken {start=} or {end=}',
                logger=self._logger
                    )
        to_split = list(self.current)
        splited = (to_split[start:end], (to_split[0:start], to_split[end:]))
        self._logger.debug(f'To arrange result: {splited}')

        return splited

    def arrange(
        self,
        arranged: List[Card],
        last: Tuple[List[Card], List[Card]]
            ) -> Deque[Card]:
        """Concatenate new current deck from given arranged list and last of
        deck. Use to_arrange() method to get list to arrange and last.

        Args:
            arranged (List[Card]): arranged list of cards
            last: (Tuple[List[Card], List[Card]]): last of deck
        """
        reorranged: Deque = deque()
        reorranged.extend(last[0])
        reorranged.extend(arranged)
        reorranged.extend(last[1])

        if len(reorranged) == len(self.current):
            self.current = reorranged
            self._logger.debug(f'Arrange result: {reorranged}')
        else:
            raise ArrangeIndexError(
                f'Wrong to_arranged parts: {arranged=}, {last=}',
                logger=self._logger
                    )
        return self.current

    def search(
        self,
        query: Dict[str, int],
        remove: bool = True
            ) -> List[Card]:
        """Search for cards in current deck

        Args:
            query (Dict[str, int]): dict with id of searched
                                    cards and count of searching
            remove (bool): if True - remove searched cards from
                           current deck. Default to True.

        Return:
            List[Card]: list of find cards

        .. code-block::
            :caption: Example:

                game.deck1.search(
                    {'card1': 2,
                     'card2': 1 },
                    remove=False
                    )
        """
        for_deque: Deque = deque()
        result = []

        while True:
            try:
                card = self.current.popleft()
                if card.id in query.keys() and query[card.id] > 0:
                    result.append(card)
                    query[card.id] -= 1
                    if not remove:
                        for_deque.append(card)
                else:
                    for_deque.append(card)
            except IndexError:
                break
        self.current = for_deque
        self._logger.debug(f'Search result: {result}')

        return result

    def get_random(
        self,
        count: int = 1,
        remove: bool = True
            ) -> List[Card]:
        """Get random cards from current deck

        Args:
            count (int, optional): count of random cards. Defaults to 1.
            remove (bool, optional): if True - remove random cards from
                                     current deck. Default to True.

        Returns:
            List[Card]: list of random cards
        """
        if not self.current:
            self._logger.debug(
                'Is empty current deck. Random cards not choosed.'
                    )
            return []
        if not remove:
            result = random.choices(self.current, k=count)
            self._logger.debug(
                f'Random choised cards without remove: {result}'
                    )
            return result
        else:
            result = []
            for _ in range(count):
                if self.current:
                    choice = random.choice(self.current)
                    result.append(choice)
                    self.current.remove(choice)
                else:
                    break
            self._logger.debug(
                f'Random choised cards with remove: {result}'
                    )
            return result

    def get_current_names(self) -> List[str]:
        """Get ids of current objects

        Returns:
            List[str]: list ids in current attribut of tool
        """
        return [obj.id for obj in self.current]


@dataclass_json
@dataclass(repr=False)
class Steps(BaseTool):
    """Game steps order object

    Attr:
        - current (List[Tuple[int, Step]]): current order of steps.
    """
    current: List[Tuple[int, Step]] = field(
        default_factory=list,
        metadata=config(exclude=lambda x: True),  # type: ignore
        repr=False,
            )

    def clear(self) -> None:
        """Clear the current queue
        """
        self.current = []

    def push(self, item: Step) -> None:
        """Push Step object to queue

        Args:
            item (Step): Step class instance
        """
        heappush(self.current, (item.priority, item))

    def pull(self) -> Step:
        """Pull Step object from queue with lowest priority

        Returns:
            Step: Step instance object
        """
        return heappop(self.current)[1]

    def _step_replace(self, step: Step) -> None:
        """Replace card in current
        """
        replaced = replace(step)
        self.push(replaced)

    def deal(
        self,
        source: Optional[List[str]] = None
            ) -> List[Tuple[int, Step]]:
        """Clear current order and create new current order

        Args:
            source (Optional[List[str]]): list of stuff ids

        Returns:
            List[Tuple[int, Step]] - list of priority and steps
        """
        self.clear()

        if not source:
            for comp in self:
                if isinstance(self[comp], Step):
                    self._step_replace(self[comp])

        else:

            names = self.get_names()
            used: Dict[str, Step] = {}

            for name in source:
                if name in names:
                    if not used.get(name):
                        for comp in self:
                            if isinstance(self[comp], Step) \
                                    and self[comp].id == name:
                                self._step_replace(self[comp])
                                break
                    else:
                        self._step_replace(self[comp])

        self._logger.debug(f'Is deal order of turn: {self.current}')
        return self.current

    def get_current_names(self) -> List[str]:
        """Get ids of current objects

        Returns:
            List[str]: list ids in current attribut of tool
        """
        return [obj[1].id for obj in self.current]



"""Game markers, like counters and etc
"""
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from bgameb.base import Base


@dataclass_json
@dataclass(repr=False)
class BaseMarker(Base):
    """Base class for game markers
    """

    def __post_init__(self) -> None:
        super().__post_init__()


@dataclass_json
@dataclass(order=True, repr=False)
class Step(BaseMarker):
    """Game steps or turns

    Attr:
        - priority (int): priority queue number. Default to 0.
    """
    priority: int = 0

    def __post_init__(self) -> None:
        super().__post_init__()



"""Types of package objects
"""
from typing import List


MARKERS: List[str] = ['step']
ITEMS: List[str] = ['dice', 'card', ]
TOOLS: List[str] = ['shaker', 'deck', 'steps']
PLAYERS: List[str] = ['player', ]
GAMES: List[str] = ['game', ]

MARKERS_ITEMS = MARKERS + ITEMS
MARKERS_ITEMS_TOOLS = MARKERS_ITEMS + TOOLS
COMPONENTS = MARKERS_ITEMS_TOOLS + PLAYERS + GAMES



"""Game dices, coins, cards and other items
"""
import random
from typing import List, Optional
from dataclasses import dataclass, field
from dataclasses_json import config, dataclass_json
from bgameb.base import Base
from bgameb.errors import StuffDefineError
from bgameb.types import MARKERS


@dataclass_json
@dataclass(repr=False)
class BaseItem(Base):
    """Base class for game items (like dices or cards)

    Attr:
        - count (int): count of items. Default to 1.

    """
    count: int = 1

    def __post_init__(self) -> None:
        super().__post_init__()
        self._types_to_add = MARKERS


@dataclass_json
@dataclass(repr=False)
class Dice(BaseItem):
    """Rolled or fliped objects, like dices or coins.

    Sides attr define number of sides of roller. Default to 2.
    Sides can't be less than 2.

    .. code-block::
        :caption: Example:

            dice = Dice('coin', sides=2)

    Raises:
        StuffDefineError: number of sides less than 2
    """
    sides: int = 2
    _range: List[int] = field(
        default_factory=list,
        metadata=config(exclude=lambda x: True),  # type: ignore
        init=False,
        repr=False
        )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.sides < 2:
            raise StuffDefineError(
                message=f'Number of sides={self.sides} '
                        f'for "{self.id}". Needed >= 2.',
                logger=self._logger
                )
        self._range = list(range(1, self.sides + 1))

    def roll(self) -> List[int]:
        """Roll and return result

        Returns:
            List[int]: result of roll
        """
        roll = [
            random.choices(self._range, k=1)[0] for _
            in list(range(self.count))
            ]
        self._logger.debug(f'Is rolled {roll=}')
        return roll


@dataclass_json
@dataclass(repr=False)
class Card(BaseItem):
    """Card object

    Attr:
        - opened (bool): is card oppened. Default to False.
        - tapped (bool): is card tapped. Default to False.
        - side (str, optional): the side of tap. Default to None.

    .. code-block::
        :caption: Example:

            card = CardType('unique_card')
            card.tap(side='left')
    """
    opened: bool = False
    tapped: bool = False
    side: Optional[str] = None

    def __post_init__(self) -> None:
        super().__post_init__()

    def flip(self) -> None:
        """Face up or face down the card regardles of it condition
        """
        if self.opened:
            self.opened = False
            self._logger.debug('Card face down.')
        else:
            self.opened = True
            self._logger.debug('Card face up.')

    def open(self) -> None:
        """Face up the card
        """
        self.opened = True
        self._logger.debug('Card face up.')

    def hide(self) -> None:
        """Face down the card
        """
        self.opened = False
        self._logger.debug('Card face down.')

    def tap(self, side='right') -> None:
        """Tap the card to the given side

        Args:
            side (str, optional): side to tap. Defaults to 'right'.
        """
        self.tapped = True
        self.side = side
        self._logger.debug(f'Card taped to side {side}.')

    def untap(self) -> None:
        """Untap the card
        """
        self.tapped = False
        self.side = None
        self._logger.debug('Card untaped. Side set to None.')

    def alter(self) -> None:
        """Many cards have alter views. For example
        card can have main view, that apply most time of the game
        and second view, that apply only if card played as
        that alternative. For ease of understanding, consider that
        different views of the same card are not related directly
        to each other.
        """
        raise NotImplementedError



"""Game players classes
"""
from typing import Optional
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from bgameb.base import Base
from bgameb.types import MARKERS_ITEMS_TOOLS


@dataclass_json
@dataclass(repr=False)
class Player(Base):
    """Base class for game players and bots

    Attr:
        - name (str): player name
        - has_priority (bool): Priority flag. Default to False
        - team (str, optioanl): team name for player. Default to None

    """
    is_active: bool = True
    has_priority: bool = False
    team: Optional[str] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        self._types_to_add = MARKERS_ITEMS_TOOLS



"""Custom error classes
"""
from __future__ import annotations
import loguru


class CustomRuntimeError(RuntimeError):
    """Base class for other runtime exceptions
    """


class ComponentIdError(CustomRuntimeError):
    """Component with given name has not wrong name or name isnt unique.
    """
    def __init__(self, id: str) -> None:
        self.message = f'{id=} is not a string.'
        super().__init__(self.message)


class ComponentNameError(CustomRuntimeError):
    """Component with given name has not wrong name or name isnt unique.
    """
    def __init__(self, name: str) -> None:
        self.message = f'Component with {name=} is exist in ' + \
                        'Component class instance or wrong name'
        super().__init__(self.message)


class ComponentClassError(CustomRuntimeError):
    """Given class isn't component.
    """
    def __init__(self, obj_, logger: loguru.Logger) -> None:
        self.message = f'Given: {obj_} not a component or you are trying ' +\
                        'to place a component in a non-designated component.'
        logger.exception(self.message)
        super().__init__(self.message)


class StuffDefineError(AttributeError):
    """Badd definition of item.
    """
    def __init__(self, message: str, logger: loguru.Logger) -> None:
        self.message = message
        logger.exception(self.message)
        super().__init__(self.message)


class ArrangeIndexError(IndexError):
    """Index error for arrange tool.
    """
    def __init__(self, message: str, logger: loguru.Logger) -> None:
        self.message = message
        logger.exception(self.message)
        super().__init__(self.message)
